#include <stdio.h>  /* printf */
#include <string.h> /* memmove */
#include <stdlib.h> /* random */

#include <math.h>  /* sin, needed just for test cases */

/* compile e.g. with: 
      gcc -Wall -O2 -lm hb.c 
*/


struct filter_env {
    int buf[16];
    int state;
    unsigned int lp;
    unsigned int sum;
    int led;
};

#define N_FILTER 8
static struct filter_env fi[N_FILTER];

void halfband(int sample, int r, struct filter_env * fbase) {
/* filtercoeff from gnu octave:

round(256*firls(14,[0 0.37 0.63 1],[1 1 0 0]))

*/
    int s,lp,hp;
    struct filter_env * f = fbase+r;

    if (r==N_FILTER-1) {
        /* stop recursion now */
        /* we've reached the end of the chain of halfband subsamplers... */
        int i;
        unsigned int tmp;
        /* add the squares (="electrical Power") & scale a little bit... */
        f->sum+=(sample*sample+16384)/32768;
        for (i=0;i<N_FILTER;i++) {
            tmp = fbase[i].sum/(1<<(N_FILTER-1-i));
            fbase[i].sum=0;
            /* do some iir lowpass filtering to get nice led bar display */
            fbase[i].lp = (fbase[i].lp*64 - fbase[i].lp + tmp + 32)/64;
            /* logarithmic scale for the leds, each led is 3dB  */
            if ((tmp=fbase[i].lp)) {
                fbase[i].led=8;
                while ((fbase[i].led) && ((tmp & 0xffffc000) == 0)) {
                    fbase[i].led--;
                    tmp <<=1;
                }
            } else {
                fbase[i].led=0;
            }
        } /* for i */
        /* now the Status of fbase[i].led is valid */
        return;
    }
    /* here starts the filtering... */
    /* halfband filtering and decimation by 2... */
    if (f->state) {
        memmove(f->buf,f->buf+2,(16-2)*sizeof(int));
        f->buf[16-2]=sample;
        /* toggle */
        f->state = 0;
    } else {
        f->buf[16-1]=sample;
        /* calculate with the filtercoeff. obtained with gnu octave */
        s = (f->buf[ 0]+f->buf[14])*(-3) +
            (f->buf[ 2]+f->buf[12])*  9  -
            (f->buf[ 4]+f->buf[10])* 22  +
            (f->buf[ 6]+f->buf[ 8])* 80;
        lp=(s+f->buf[7]*128)/256;
        hp=(s-f->buf[7]*128)/256;

        /* add the squares of the high-pass (="electrical Power") & scale a little bit... */
        f->sum+=(hp*hp+16384)/32768;
        /* recursively call the "next" halfband filter... */
        halfband( lp, r+1, fbase);
        /* toggle */
        f->state = 1;
    }
}

int main()
{
  int i;
  int sample=0;
  for (i=0;i<22050;i++)
  {
    /* white noise */
    sample=(random()& 0xffff)-32768;
    /* sawtooth */
//    sample=(i%64)*1024-32768+512;
    /* sine */
//    sample=(int)(32767.0*sin(i/192.0*M_PI*2.0));

    halfband(sample, 0, &fi[0]);
  }
  for (i=0;i<N_FILTER;i++) {
     printf("%d\n",fi[i].led);
  }
  return 0;
}
